/* ftp.c
   File Transfer Protocol support module */

#include "config.h"

#include <stdio.h>
#ifdef HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include "geturl.h"
#include "ftp.h"
#include "connect.h"
#include "host.h"
#include "http.h"

extern char *ftp_acc, *ftp_pass;
extern int verbose, quiet, debug;
extern FILE *lfile;

extern int errno;

char respline[MAX_RESPLINE];


/* Write the response of FTP server in line, consisting of not more
   than maxlen bytes. <CR> and <LF> characters are stripped from the
   line. The lines that have '-' as their fourth characters are
   skipped, until the line without it is found. The line is
   null-terminated. */
uerr_t
ftp_response(int fd, char *line, int maxlen)
{
   int i, res;

   do
   {
      for (i = 0; i < maxlen; i++)
      {
	 res = iread(fd, line + i, 1);
	 /* res is number of bytes read. */
	 if (res == 1)
	 {
	    if (line[i] == '\n')
	    {
	       line[i] = 0;
	       /* Get rid of \r. */
	       if (i > 0 && line[i - 1] == '\r')
		  line[i - 1] = 0; /* He he */
	       break;
	    }
	 }
	 else
	    return FTPRERR;
      }
#ifdef DEBUG
      if (debug)
      {
	 fprintf(lfile, "%s\n", line);
	 fflush(lfile);
      }
#endif
      if (line[i])
	 return FTPRERR;
   } while (line[3] == '-');
   return FTPOK;
}

/* Returns the malloc-ed FTP request, ending with <CR><LF>. */
char *
ftp_request(char *command, char *value)
{
   char *res;

   res = nmalloc(strlen(command) + 1 + strlen(value) + 2 + 1);
   sprintf(res, "%s %s\r\n", command, value);
#ifdef DEBUG
   if (debug)
   {
      fprintf(lfile, "\n--> %s %s\n", command, value);
      fflush(lfile);
   }
#endif
   return res;
}

/* Sends the USER and PASS commands to the server, to control
   connection socket csock. */
uerr_t
ftp_login(int csock, char *acc, char *pass)
{
   uerr_t err;
   char *request;

   /* Get greeting. */
   err = ftp_response(csock, respline, MAX_RESPLINE - 1);
   if (err != FTPOK)
      return err;
   if (*respline != '2')
      return FTPSRVERR;
   /* Send USER username. */
   request = ftp_request("USER", acc);
   write(csock, request, strlen(request));
   free(request);
   /* Get appropriate response. */
   err = ftp_response(csock, respline, MAX_RESPLINE - 1);
   if (err != FTPOK)
      return err;
   /* An unprobable possibility of logging without a password. */
   if (*respline == '2')
      return FTPOK;
   /* Else, only response 3 is appropriate. */
   if (*respline != '3')
      return FTPLOGINC;
   /* Send PASS password. */
   request = ftp_request("PASS", pass);
   write(csock, request, strlen(request));
   free(request);
   /* Get appropriate response. */
   err = ftp_response(csock, respline, MAX_RESPLINE - 1);
   if (err != FTPOK)
      return err;
   if (*respline != '2')
      return FTPLOGINC;
   /* All OK. */
   return FTPOK;
}

/* Binds a port, and sends the appropriate PORT command to the FTP
   server. Use acceptport after RETR, to get the socket of data
   connection. */
uerr_t
ftp_port(int csock)
{
   uerr_t err;
   char *request, *bytes;
   unsigned char *in_addr;
   int port;

   /* Setting port to 0 lets the system choose a free port. */
   port = 0;
   /* Bind the port. */
   err = bindport(&port);
   if (err != BINDOK)
      return err;
   /* Get my INET address. */
   if (!(in_addr = inet_address()))
      return HOSTERR;
   /* Construct the argument of PORT (of the form a,b,c,d,e,f) */
   bytes = nmalloc(6 * 4 + 1);
   sprintf(bytes, "%d,%d,%d,%d,%d,%d", in_addr[0], in_addr[1],
	   in_addr[2], in_addr[3], (port & 0xff00) >> 8, port & 0xff);
   /* Send PORT request. */
   request = ftp_request("PORT", bytes);
   free(bytes);
   write(csock, request, strlen(request));
   free(request);
   /* Get appropriate response. */
   err = ftp_response(csock, respline, MAX_RESPLINE - 1);
   if (err != FTPOK)
      return err;
   if (*respline != '2')
      return FTPPORTERR;
   return FTPOK;
}

/* Sends the TYPE request to the server. */
uerr_t
ftp_type(int csock, int type)
{
   char *request;
   uerr_t err;
   char stype[2];

   /* Construct argument. */
   stype[0] = type;
   stype[1] = 0;
   /* Send TYPE request. */
   request = ftp_request("TYPE", stype);
   write(csock, request, strlen(request));
   free(request);
   /* Get appropriate response. */
   err = ftp_response(csock, respline, MAX_RESPLINE - 1);
   if (err != FTPOK)
      return err;
   if (*respline != '2')
      return FTPUNKNOWNTYPE;
   /* All OK. */
   return FTPOK;
}

/* Changes the working directory by issuing a CWD command to the
   server. */
uerr_t
ftp_cwd(int csock, char *dir)
{
   char *request;
   uerr_t err;

   /* Send CWD request. */
   request = ftp_request("CWD", dir);
   write(csock, request, strlen(request));
   free(request);
   /* Get appropriate respone. */
   err = ftp_response(csock, respline, MAX_RESPLINE - 1);
   if (err != FTPOK)
      return err;
   if (*respline == '5')
      return FTPNSFOD;
   if (*respline != '2')
      return FTPRERR;
   /* All OK. */
   return FTPOK;
}

/* Sends REST command to the FTP server. */
uerr_t
ftp_rest(int csock, long offset)
{
   char *request;
   uerr_t err;
   static char buf[20]; /* Buffer for the number */

   sprintf(buf, "%ld", offset);
   request = ftp_request("REST", buf);
   write(csock, request, strlen(request));
   free(request);
   /* Get appropriate response. */
   err = ftp_response(csock, respline, MAX_RESPLINE - 1);
   if (err != FTPOK)
      return err;
   if (*respline != '3')
      return FTPRESTFAIL;
   /* All OK. */
   return FTPOK;
}

/* Sends RETR command to the FTP server. */
uerr_t
ftp_retr(int csock, char *file)
{
   char *request;
   uerr_t err;

   /* Send RETR request. */
   request = ftp_request("RETR", file);
   write(csock, request, strlen(request));
   free(request);
   /* Get appropriate response. */
   err = ftp_response(csock, respline, MAX_RESPLINE - 1);
   if (err != FTPOK)
      return err;
   if (*respline == '5')
      return FTPNSFOD;
   if (*respline != '1')
      return FTPRERR;
   /* All OK. */
   return FTPOK;
}

/* Sends the LIST command to the server. */
uerr_t
ftp_list(int csock, char *file)
{
   char *request;
   uerr_t err;

   /* Send LIST request. */
   request = ftp_request("LIST", file);
   write(csock, request, strlen(request));
   free(request);
   /* Get appropriate respone. */
   err = ftp_response(csock, respline, MAX_RESPLINE - 1);
   if (err != FTPOK)
      return err;
   if (*respline == '5')
      return FTPNSFOD;
   if (*respline != '1')
      return FTPRERR;
   /* All OK. */
   return FTPOK;
}

/* Retrieves a file with denoted parameters through opening an FTP
   connection to the server. Control and data socket are closed
   afterwards. */
uerr_t
getftp(char *remote_host, int remote_port, char *remote_file,
       char *local_file, long *len, long restval,
       int *res, int *type)
{
   int csock, dtsock, i, list_only = 0;
   uerr_t err;
   char *dir, *file, *buf;
   FILE *fp;

   *type &= ~HTTP200;
   *len = 0;
   /* First, some checking... */
   buf = nstrdup(remote_file);
   for (i = strlen(buf); i && buf[i] != '/'; i--);
   if (!i && buf[i + 1]) /* Just filename or /filename */
   {
      dir = "/";
      file = buf + (*buf == '/' ? 1 : 0);
   }
   else if (!i) /* Just dirname */
   {
      dir = buf;
      file = "";
   }
   else /* Everything OK */
   {
      buf[i] = 0;
      dir = buf;
      file = buf + i + 1;
   }
   if (!strlen(file)) /* Empty file */
      list_only = 1;
   if (!strlen(dir))
      dir = "/";
   /* Then, let's establish the connection. */
   if (verbose)
   {
      fprintf(lfile, "Connecting to %s:%d... ", remote_host, remote_port);
      fflush(lfile);
   }
   err = make_connection(&csock, remote_host, remote_port);
   switch (err)
   {
      /* Do not close the socket in first several cases,
         since it wasn't made in the first place. */
      case HOSTERR:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "No such host %s.\n", remote_host);
	    fflush(lfile);
	 }
	 free(buf);
	 return HOSTERR;
	 break;
      case CONSOCKERR:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "Socket error (%s).\n", mystrerror(errno));
	    fflush(lfile);
	 }
	 free(buf);
	 return CONSOCKERR;
	 break;
      case CONTIMEDOUT:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "Connection to %s timed out.\n", remote_host);
	    fflush(lfile);
	 }
	 free(buf);
	 close(csock);
	 return CONTIMEDOUT;
	 break;
      case CONREFUSED:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "Connection to %s refused.\n", remote_host);
	    fflush(lfile);
	 }
	 free(buf);
	 close(csock);
	 return CONREFUSED;
      case CONERROR:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "connect: %s\n", mystrerror(errno));
	    fflush(lfile);
	 }
	 free(buf);
	 close(csock);
	 return CONERROR;
	 break;
   }
   /* Log in. */
   if (verbose)
   {
      fprintf(lfile, "Connected!\n");
      fprintf(lfile, "Logging in as %s ... ", ftp_acc);
      fflush(lfile);
   }
   err = ftp_login(csock, ftp_acc, ftp_pass);
   /* All FTPRERR, FTPSRVERR and FTPLOGININC. */
   switch (err)
   {
      case FTPRERR:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile,
		    "Error in server response, control connection closed.\n");
	    fflush(lfile);
	 }
	 free(buf);
	 close(csock);
	 return err;
	 break;
      case FTPSRVERR:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "Error in server greeting.\n");
	    fflush(lfile);
	 }
	 free(buf);
	 close(csock);
	 return err;
      case FTPLOGINC:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "Login incorrect or the server refuses to connect.\n");
	    fflush(lfile);
	 }
	 free(buf);
	 close(csock);
	 return FTPLOGINC;
	 break;
      case FTPOK:
	 if (verbose)
	 {
	    fprintf(lfile, "Logged in!\n");
	    fflush(lfile);
	 }
	 break;
      default:
	 fprintf(lfile, "This condition should not have been reached.\n");
	 fflush(lfile);
	 exit(1);
	 break;
   }
   /* Change type to binary. */
   if (verbose)
   {
      fprintf(lfile, "==> TYPE I ... ");
      fflush(lfile);
   }
   err = ftp_type(csock, 'I');
   /* ftp_response + FTPUNKNOWNTYPE. */
   switch (err)
   {
      case FTPRERR:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile,
		    "Error in server response, control connection closed.\n");
	    fflush(lfile);
	 }
	 free(buf);
	 close(csock);
	 return err;
	 break;
      case FTPUNKNOWNTYPE:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "Unknown type.\n");
	    fflush(lfile);
	 }
	 free(buf);
	 close(csock);
	 return err;
      default:
	 /* Everything is OK. */
	 break;
   }
   /* Change working directory. */
   if (verbose)
   {
      fprintf(lfile, "done!  ");
      fprintf(lfile, "==> CWD %s ... ", dir);
      fflush(lfile);
   }
   err = ftp_cwd(csock, dir);
   /* Possible errors: ftp_response + FTPNSFOD. */
   switch (err)
   {
      case FTPRERR:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile,
		    "Error in server response, control connection closed.\n");
	    fflush(lfile);
	 }
	 free(buf);
	 close(csock);
	 return err;
	 break;
      case FTPNSFOD:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "No such directory %s.\n", dir);
	    fflush(lfile);
	 }
	 free(buf);
	 close(csock);
	 return err;
	 break;
   }
   /* Send PORT. */
   if (verbose)
   {
      fprintf(lfile, "done!\n");
      fprintf(lfile, "==> PORT ... ");
      fflush(lfile);
   }
   err = ftp_port(csock);
   /* All possible errors from bindport and ftp_response
      plus FTPPORTERR and HOSTERR. */
   switch(err)
   {
      case HOSTERR:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "No such host %s.\n", remote_host);
	    fflush(lfile);
	 }
	 free(buf);
	 close(csock);
	 return HOSTERR;
	 break;
      case CONSOCKERR:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "Socket error (%s).\n", mystrerror(errno));
	    fflush(lfile);
	 }
	 free(buf);
	 close(csock);
	 return CONSOCKERR;
	 break;
      case BINDERR: case CONPORTERR: case LISTENERR:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "Bind error (%s).\n", mystrerror(errno));
	    fflush(lfile);
	 }
      case FTPPORTERR:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "Invalid PORT (%s).\n", respline);
	    fflush(lfile);
	 }
      default:
	 /* Everything is OK. */
	 break;
   }
   /* Restart if needed. */
   if (restval && !list_only)
   {
      if (verbose)
      {
	 fprintf(lfile, "done!    ");
	 fprintf(lfile, "==> REST %ld ... ", restval);
      }
      err = ftp_rest(csock, restval);
      /* Possible errors: ftp_response + FTPRESTFAIL. */
      switch (err)
      {
	 case FTPRERR:
	    if (!quiet)
	    {
	       if (verbose)
		  fprintf(lfile, "\n");
	       fprintf(lfile,
		       "Error in server response, control connection closed.\n");
	       fflush(lfile);
	    }
	    free(buf);
	    close(csock);
	    return err;
	    break;
	 case FTPRESTFAIL:
	    if (verbose)
	    {
	       fprintf(lfile, "\nREST failed, starting from scratch.\n");
	       fflush(lfile);
	    }
	    restval = 0;
	    break;
      }
   }
   /* Tell the server what to send. */
   if (!list_only)
   {
      if (verbose)
      {
	 if (err != FTPRESTFAIL)
	    fprintf(lfile, "done!    ");
	 if (restval)
	    fprintf(lfile, "\n");
	 fprintf(lfile, "==> RETR %s ... ", file);
	 fflush(lfile);
      }
      err = ftp_retr(csock, file);
      /* Possible errors: ftp_response + FTPNSFOD. */
      switch (err)
      {
	 case FTPRERR:
	    if (!quiet)
	    {
	       if (verbose)
		  fprintf(lfile, "\n");
	       fprintf(lfile,
		       "Error in server response, control connection closed.\n");
	       fflush(lfile);
	    }
	    free(buf);
	    close(csock);
	    return err;
	    break;
	 case FTPNSFOD:
	    if (!quiet)
	    {
	       if (verbose)
		  fprintf(lfile, "\n");
	       fprintf(lfile, "No such file %s.\n", file);
	       fflush(lfile);
	    }
	    free(buf);
	    close(csock);
	    return err;
	    break;
      }
   }
   else
   {
      if (verbose)
      {
	 if (err != FTPRESTFAIL)
	    fprintf(lfile, "done!    ");
	 if (restval)
	    fprintf(lfile, "\n");
	 fprintf(lfile, "==> LIST ... ");
	 fflush(lfile);
      }
      err = ftp_list(csock, ".");
      /* Possible errors: ftp_response */
      switch (err)
      {
	 case FTPRERR:
	    if (!quiet)
	    {
	       if (verbose)
		  fprintf(lfile, "\n");
	       fprintf(lfile,
		       "Error in server response, control connection closed.\n");
	       fflush(lfile);
	    }
	    free(buf);
	    close(csock);
	    return err;
	    break;
	 case FTPNSFOD:
	    if (!quiet)
	    {
	       if (verbose)
		  fprintf(lfile, "\n");
	       fprintf(lfile, "No such file %s.\n", file);
	       fflush(lfile);
	    }
	    free(buf);
	    close(csock);
	    return err;
	    break;
      }
   }
   free(buf); /* Safe to do so now. */
   if (verbose)
   {
      fprintf(lfile, "done!\n");
      fflush(lfile);
   }
   /* Open the data transmission socket by calling acceptport. */
   err = acceptport(&dtsock);
   /* Possible errors: ACCEPTERR. */
   if (err == ACCEPTERR)
   {
      if (!quiet)
      {
	 fprintf(lfile, "accept: %s\n", mystrerror(errno));
	 fflush(lfile);
      }
      close(csock);
      return err;
   }
   /* Open the file... */
   fp = fopen(local_file, restval ? "a" : "w");
   if (!fp)
   {
      if (!quiet)
      {
	 fprintf(lfile, "fopen: %s\n", mystrerror(errno));
	 fflush(lfile);
      }
      close(csock);
      return FOPENERR;
   }
   /* As soon as we got away with accept, the document is considered
      HTTP200 ok. */
   *type |= HTTP200;
   
   /* Get the contents of the document. */
   *res = get_contents(dtsock, fp, len, restval);
   /* Close data connection socket. */
   closeport(dtsock);
   /* Close the local file. */
   fclose(fp);
   /* If get_contents couldn't write to fp, bail out. */
   if (*res == -2)
   {
      close(csock);
      return FWRITEERR;
   }
   /* Get server to tell you if everything is retrieved. */
   err = ftp_response(csock, respline, MAX_RESPLINE - 1);
   if (err != FTPOK)
   {
      if (!quiet)
      {
	 /* A different message here, since the connection was
            probably lost, and there is no other mysterious error in
            server response. Newline is not printed because of
            possible Data connection error. */
	 fprintf(lfile, "Control connection closed.");
	 fflush(lfile);
      }
      close(csock);
      return err;
   }
   /* If retrieval failed for any reason, return FTPRETRINT. */
   if (*respline != '2')
   {
      close(csock);
      return FTPRETRINT;
   }
   /* I should probably send 'QUIT', but this is faster. */
   close(csock);
   /* Now everything is definitely supposed to be OK. */
   if (list_only)
   {
      char **files, **dirs;
      int i;
      
      err = ftp_parse_list(local_file, &dirs, &files);
      if (err != FOPENERR)
      {
	 err = ftp_index(local_file, dirs, files, remote_host,
			 remote_file, remote_port);
	 if (err != FOPENERR)
	    if (verbose)
	    {
	       fprintf(lfile, "Wrote HTML index to %s.\n", local_file);
	       fflush(lfile);
	    }
         for (i = 0; files[i]; i++)
	    free(files[i]);
         free(files);
         for (i = 0; dirs[i]; i++)
	    free(dirs[i]);
         free(dirs);
      }
   }
   return RETRFINISHED;
}

/* The function parses the listing given in file, and returns hash
   arrays of files and directories. */
uerr_t
ftp_parse_list(char *file, char ***dirs, char ***files)
{
   FILE *fp;
   static char *months[] = {
      "Jan", "Feb", "Mar", "Apr", "May", "Jun",
      "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
   };
   int c, next, len, i;
   char *line, *tok, ***cur;

   fp = fopen(file, "r");
   if (!fp)
      return FOPENERR;
   /* Initialize dirs and files. */
   *dirs = nmalloc(sizeof(char **));
   *files = nmalloc(sizeof(char **));
   (*dirs)[0] = (*files)[0] = NULL;

   /* Line loop to end of file: */
   do
   {
      line = nmalloc(1);
      line[0] = 0;
      len = 0;
      /* Character loop to end of line: */
      do
      {
	 c = fgetc(fp);
	 /* Exit on EOF or newline. */
	 if (c == EOF || c == '\n')
	    break;
	 /* Increase the line siye. */
	 line = nrealloc(line, len + 2);
	 line[len] = c;
	 line[len + 1] = 0;
	 ++len;
      } while (1);
      /* Destroy <CR> if there is one. */
      if (line[len] == '\r')
	 line[len] = 0;
      /* Skip if total... */
      if (!strncmp(line, "total", 5))
      {
	 free(line);
	 continue;
      }
      /* Get the first token (permissions). */
      tok = strtok(line, " ");
      if (!tok)
      {
	 free(line);
	 continue;
      }
      /* Decide whether we deal with a file or a directory. */
      if (*tok == 'd')
	 cur = dirs;
      else
	 cur = files;
      
      next = -1;
      /* While there are tokens on the line: */
      while ((tok = strtok(NULL, " ")))
      {
	 --next;
	 for (i = 0; i < 12; i++)
	    if (!strcmp(tok, months[i]))
	       break;
	 /* If we got a month, it means the filename is three tokens
            away. */
	 if (i != 12)
	    next = 3;
	 /* If we have the filename, add it to the list of files or
            directories. */
	 if (!next)
	 {
	    /* "." and ".." are an exception! */
	    if (!strcmp(tok, ".") || !strcmp(tok, ".."))
	       break;
	    for (i = 0; (*cur)[i]; i++);
	    *cur = nrealloc(*cur, (i + 2)*sizeof(char *));
	    (*cur)[i] = nstrdup(tok);
	    (*cur)[i + 1] = NULL;
	 }
      }
      /* Free the line. */
      free(line);
   } while (c != EOF);
   /* Close the file. */
   fclose(fp);
   return FTPOK;
}

/* The function creates an HTML index containing references to given
   directories and files on the appropriate host. The references are
   FTP. */
uerr_t
ftp_index(char *file, char **dirs, char **files, char *host, char *dir,
	  int port)
{
   FILE *fp;
   int i;
   char *tmp;

   fp = fopen(file, "w");
   if (!fp)
      return FOPENERR;
   fprintf(fp, "<html>\n<head>\n<title>Index of ftp://%s:%d%s</title>\n</head>\n",
           host, port, dir);
   fprintf(fp, "<body>\n<h1>Index of ftp://%s:%d%s</h1>\n<hr>\n",
           host, port, dir);
   fprintf(fp, "<ul>\n");
   for (i = 0; files[i]; i++)
   {
      tmp = nmalloc(6 + strlen(host) + 1 + numdigit(port) + strlen(dir)
		    + strlen(files[i]) + 1);
      sprintf(tmp, "ftp://%s:%d%s%s", host, port, dir, files[i]);
      fprintf(fp, "   <li>File <a href=\"%s\">%s</a>\n", tmp, files[i]);
      free(tmp);
   }
   for (i = 0; dirs[i]; i++)
   {
      tmp = nmalloc(6 + strlen(host) + 1 + numdigit(port) + strlen(dir)
		    + strlen(dirs[i]) + 1 + 1);
      sprintf(tmp, "ftp://%s:%d%s%s/", host, port, dir, dirs[i]);
      fprintf(fp, "   <li>Directory <a href=\"%s\">%s</a>\n", tmp, dirs[i]);
      free(tmp);
   }
   fprintf(fp, "</ul>\n</body>\n</html>\n");
   fclose(fp);
   return FTPOK;
}

