/* main.c
   Main module, used for option parsing */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif /* HAVE_UNISTD_H */
#ifdef HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif /* HAVE_UNISTD_H */
#include "geturl.h"
#include "main.h"
#include "loop.h"
#include "init.h"
#include "getopt.h"

extern int verbose, quiet, ntry, recursive;
extern char *lfilename, *input_filename;

FILE *lfile;         /* Global log file */
extern char *version_string;

#ifdef DEBUG
extern int debug;       /* Debugging on/off */
#endif /* DEBUG */

int
main(int argc, char **argv)
{
   char **url, **t;
   char *usage =
      "Usage: %s [options] [URL-list]\nInvoke with --help to get help.\n";
   int i, c, nurl, status, append_to_log;
   char *exec_name;

   static struct option long_options[] =
   {
      { "help", 0, 0, 'h' },
      { "version", 0, 0, 'V' },
      { "prefix-files", 0, 0, 'p' },
      { "quiet", 0, 0, 'q' },
      { "verbose", 0, 0, 'v' },
      { "debug", 0, 0, 'd' },
      { "recursive", 0, 0, 'r' },
      { "span-hosts", 0, 0, 'H' },
      { "relative", 0, 0, 'L' },
      { "follow-ftp", 0, 0, 'f' },
      { "continue-ftp", 0, 0, 'c' },
      { "proxy", 0, 0, 'Y' },
      { "use-proxy", 0, 0, 'Y' },
      { "force-html", 0, 0, 'F'},
      { "level", 1, 0, 'l' },
      { "no", 1, 0, 'n' },
      { "input-file", 1, 0, 'i' },
      { "output-file", 1, 0, 'o' },
      { "append-output", 1, 0, 'a' },
      { "tries", 1, 0, 't' },
      { "domains", 1, 0, 'D' },
      { "accept", 1, 0, 'A' },
      { "reject", 1, 0, 'R' },
      { "directory-prefix", 1, 0, 'P' },
      { "timeout", 1, 0, 'T' },
      { "base", 1, 0, 'B' },
      { 0, 0, 0, 0 }
   };

   append_to_log = 0;
   initialize();
   for (i = strlen(argv[0]); i && argv[0][i] != '/'; i--);
   exec_name = argv[0] + i + (argv[0][i] == '/');
   
   while ((c = getopt_long(argc, argv, "hVpqvdrHLfcYFl:n:i:o:a:t:D:A:R:P:T:B:",
			   long_options, (int *)0)) != EOF)
   {
      switch(c)
      {
	 case 'h':
	    printhelp();
	    exit(0);
	    break;
	 case 'V':
	    printf("%s\n", version_string);
	    exit(0);
	    break;
	 case 'Y':
	    setval("useproxy", "on");
	    break;
	 case 'F':
	    setval("forcehtml", "on");
	    break;
	 case 'p':
	    setval("prefixfiles", "on");
	    break;
	 case 'q':
	    setval("quiet", "on");
	    break;
	 case 'v':
	    setval("verbose", "on");
	    break;
	 case 'l':
	    setval("reclevel", optarg);
	    break;
	 case 'd':
	    setval("debug", "on");
	    break;
	 case 'n':
	    for (i = 0; i < strlen(optarg); i++)
	       switch(optarg[i])
	       {
		  case 'v':
		     setval("verbose", "off");
		     break;
		  case 'h':
		     setval("simplehostcheck", "on");
		     break;
		  case 'H':
		     setval("addhostdir", "off");
		     break;
		  case 'd':
		     setval("dirstruct", "off");
		     break;
		  case 'c':
		     setval("noclobber", "on");
		     break;
		  default:
		     printf("-n%c option is unknown. Try %s -h.\n", optarg[i],
			    exec_name);
		     exit(1);
	       }
	    break;
	 case 'i':
	    setval("input", optarg);
	    break;
	 case 'r':
	    setval("recursive", "on");
	    break;
	 case 'H':
	    setval("spanhosts", "on");
	    break;
	 case 'L':
	    setval("relativeonly", "on");
	    break;
	 case 'f':
	    setval("followftp", "on");
	    break;
	 case 'c':
	    setval("alwaysrest", "on");
	    break;
	 case 'o':
	    setval("logfile", optarg);
	    break;
	 case 'a':
	    setval("logfile", optarg);
	    append_to_log = 1;
	    break;
	 case 't':
	    setval("numtries", optarg);
	    break;
	 case 'D':
	    setval("domains", optarg);
	    break;
	 case 'A':
	    setval("accept", optarg);
	    break;
	 case 'R':
	    setval("reject", optarg);
	    break;
	 case 'P':
	    setval("dirprefix", optarg);
	    break;
	 case 'T':
	    setval("timeout", optarg);
	    break;
	 case 'B':
	    setval("basehref", optarg);
	    break;
	 case '?':
	    printf(usage, exec_name);
	    exit(0);
	    break;
      }
   }
   if (verbose == -1)
      verbose = !quiet;
   if (verbose && quiet)
   {
      printf("Can't be verbose and quiet at the same time.\n");
      printf(usage, exec_name);
      exit(1);
   }
   /* Allocate basic pointer. */
   url = nmalloc(sizeof(char *));
   url[0] = NULL;
   /* Fill in the arguments. */
   for (nurl = 0; optind < argc; optind++, nurl++)
   {
      url = nrealloc(url, (nurl + 2) * sizeof(char *));
      url[nurl] = nstrdup(argv[optind]);
      url[nurl + 1] = NULL;
   }
   if (!url[0] && !input_filename)
   {
      printf("No URL specified.\n");
      printf(usage, exec_name);
      exit(1);
   }
   /* Open log filename. */
   if (lfilename)
   {
      char *s;

      s = append_to_log ? "a" : "w";
      lfile = fopen(lfilename, s);
      if (lfile == NULL)
      {
	 perror("log file");
	 exit(1);
      }
   }
   else
      lfile = stdout;

   /* Call an appropriate function. */
   if (recursive)
      status = recursive_retrieve(url, input_filename, ntry, 1, NULL);
   else
      if (!input_filename)
	 status = retrieve_more_urls(url, ntry);
      else
	 status = retrieve_urls_from_file(url, input_filename, ntry);

   /* Close the log file. */
   fclose(lfile);

   /* Free the data. */
   for (t = url; *t; t++)
      free(*t);
   free(url);
   freemem();
   return !status;
}

/* Print the help message. */
void
printhelp(void)
{
   printf("---=== %s by Hrvoje Niksic <hniksic@srce.hr> ===---\n\n",
	  version_string);
   printf("Options:\n");
   printf("-h --help             print this help message\n");
   printf("-v --verbose          verbose\n");
   printf("-q --quiet            quiet (no output)\n");
   printf("-i --input-file       read URL-s from file (use \"-\" for stdin)\n");
   printf("-o --output-file      log messages to logfile\n");
   printf("-t --tries            set number of retries per URL (0 for infinity)\n");
   printf("-nH --no=H            do not create hostname directories\n");
   printf("-r --recursive        recursive web-suck -- use with care!\n");
   printf("-l --level            maximum recursion depth (0 for infinity)\n");
   printf("-L --relative-only    follow relative links only\n");
   printf("-D --domains          comma-separated list of domains to accept\n");
   printf("-P --directory-prefix top of directory tree\n");
   printf("-A --accept           comma-separated list of extensions to accept\n");
   printf("-R --reject           comma-separated list of extensions to reject\n");
   printf("-f --follow-ftp      follow FTP links from HTML documents\n");
   printf("-c --continue-ftp    always send REST\n\n");
#ifdef NO_RECURSION
   printf("note: recursion disabled!\n\n");
#endif
   printf("Refer to the manual page for a complete list of options.\n");
   printf("Mail all bug reports to hniksic@srce.hr\n");
}

/* nmalloc, nrealloc and nstrdup exit the program if there is not
   enough memory. nstrdup also implements strdup on systems that
   do not have it. */
void *
nmalloc(size_t size)
{
   void *res;

   res = malloc(size);
   if (!res)
      memfatal("malloc");
   return res;
}

void *
nrealloc(void *obj, size_t size)
{
   void *res;

   /* Not all Unixes have the feature of realloc() that calling it
      with a NULL-pointer is the same as malloc(), but it is easy to
      simulate. */
   if (obj)
      res = realloc(obj, size);
   else
      res = malloc(size);
   if (!res)
      memfatal("realloc");
   return res;
}

char *
nstrdup(char *s)
{
   char *s1;

#ifndef HAVE_STRDUP
   int l;
   
   l = strlen(s);
   s1 = malloc(l + 1);
   if (!s1)
      memfatal("strdup");
   memcpy(s1, s, l + 1);
   return s1;
#else
   s1 = strdup(s);
   if (!s1)
      memfatal("strdup");
   return s1;
#endif
}

void
memfatal(char *s)
{
   if (!quiet)
   {
      fprintf(lfile, "%s: Not enough memory.\n", s);
      fflush(lfile);
   }
   exit(1);
}

/* Calls internal getopt function to enable long option names. */
int
getopt_long (int argc, char *const *argv, const char *shortopts,
	     const struct option *longopts, int *longind)
{
  return _getopt_internal (argc, argv, shortopts, longopts, longind, 0);
}
