/* url.c
   Module for URL handling */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#ifdef HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif
#include "geturl.h"
#include "url.h"
#include "host.h"

extern int verbose, debug, prefix_files, simple_check;
extern int dirstruct, dirmode, add_hostdir, recursive, always_rest;
extern char *dir_prefix;
extern FILE *lfile;

extern char **accepts, **rejects;
extern char *http_proxy, *ftp_proxy;

extern int errno;

/* NULL-terminated list of recognized prototypes. Note that
   recognized doesn't mean supported -- only HTTP is supported
   for now. */
char *protostrings[] = {
   "http://",
   "ftp://",
   "mailto:",
   "news:",
   "news://",
   "telnet://",
   "wais://",
   "file://",
   "gopher://",
   NULL
};

/* Similar as former, but for supported protocols: */
proto_t sup_protos[] = {
   { "http://", URLHTTP, DEFAULT_HTTP_PORT },
   { "ftp://", URLFTP, DEFAULT_FTP_PORT },
   { NULL, FTPOK, 0 }
};

/* Returns the proto-type if it is a supported protocol, or URLUNKNOWN
   if not. */
uerr_t
urlproto(char *url)
{
   int i;

   for (i = 0; sup_protos[i].name; i++)
      if (!strncasecmp(url, sup_protos[i].name, strlen(sup_protos[i].name)))
	 return sup_protos[i].ind;
   for (i = 0; url[i] && url[i] != ':' && url[i] != '/'; i++);
   if (url[i] == ':')
   {
      for (++i; url[i] && url[i] != '/'; i++)
	 if (!isdigit(url[i]))
	    return URLBADPORT;
      if (url[i - 1] == ':')
	 return URLFTP;
      else
	 return URLHTTP;
   }
   else
      return URLHTTP;
}

/* Extract the given URL of the form (http:|ftp:)//hostname(:port)?/dirname
   1. hostname (terminated with '/' or ':')
   2. port number (terminated with '/'), or chosen for the protocol
   3. dirname (everything after hostname, '/' if hostname is unterminated
   Most errors are handled. No allocation is done, you must supply
   pointers to allocated memory.

   Later: Added recognizing of:
   hostname:/dir/file          for FTP and
   hostname(:portnum)/dir/file for HTTP.  */
uerr_t
parseurl(char *url, char *hostname, int *port, char *dirname)
{
   int i, l, recognizable, changed_port = 0;
   uerr_t type;

   for (i = 0; protostrings[i]; i++)
      if (!strncasecmp(protostrings[i], url, strlen(protostrings[i])))
	 break;
   /* If proto is found, it is recognizable. */
   recognizable = protostrings[i] ? 1 : 0;
   for (i = 0; sup_protos[i].name; i++)
   {
      l = strlen(sup_protos[i].name);
      if (!strncasecmp(sup_protos[i].name, url, l))
         break;
   }
   /* If protocol is recognizable, but unsupported, bail out, else
      suppose unknown. */
   if (recognizable && !sup_protos[i].name)
      return URLUNKNOWN;
   else if (!sup_protos[i].name)
      type = URLUNKNOWN;
   else
      type = sup_protos[i].ind;
   
   if (type == URLUNKNOWN)
      l = 0;
   for (i = l; url[i] && url[i] != ':' && url[i] != '/'; i++)
      hostname[i - l] = url[i];
   hostname[i - l] = 0;
   if (!*hostname)
      return URLBADHOST;
   *port = 0;
   changed_port = 0;
   if (url[i] == ':')
   {
      changed_port = 1;
      for (++i; url[i] && url[i] != '/'; i++)
	 if (isdigit(url[i]))
	 {
	    *port *= 10;
	    *port += url[i] - '0';
	 }
	 else
	    return URLBADPORT;
      if (type == URLUNKNOWN)
      {
	 if (url[i - 1] == ':')
	 {
	    type = URLFTP;
	    changed_port = 0;
	 }
	 else
	    type = URLHTTP;
      }
      else if (url[i - 1] == ':')
	 return URLBADPORT;
   }
   else if (type == URLUNKNOWN)
      type = URLHTTP;
   if (!url[i])
      strcpy(dirname, "/");
   else
      strcpy(dirname, url + i);
   if (!*port)
   {
      if (changed_port)
	 return URLBADPORT;
      for (i = 0; sup_protos[i].name; i++)
	 if (sup_protos[i].ind == type)
	    break;
      if (!sup_protos[i].name)
	 return URLUNKNOWN;
      *port = sup_protos[i].port;
   }
   return type;
}


/* Find URL of format
   (http://|ftp://|mailto:|(...))hostname[:port]/dir -- no HTML
   parsing done, it can be any text/binary buffer */
char *
findurl(char *buf, int howmuch, int *count)
{
   char **prot, *s1, *s2;

   for (s1 = buf; howmuch; s1++, howmuch--)
      for (prot = protostrings; *prot; prot++)
	 if (howmuch <= strlen(*prot))
	    break;
	 else if (!strncasecmp(*prot, s1, strlen(*prot)))
	 {
	    for (s2 = s1, *count = 0;
		 *s2 && *s2 >= 32 && *s2 < 127 && !isspace(*s2) &&
		    !strchr(URL_SEPARATOR, *s2);
		 s2++, (*count)++);
	    return s1;
	 }
   return NULL;
}

/* 1. Parse the given URL, to extract hostname, dirname and filename
   2. Make the required hierarchy of directories, as with mkdir -p
   3. Return the pathname of the URL-equivalent filename */
char *
mkstruct(char *url)
{
   char *host, *nhost, *odir, *dir, *res;
   int port, pos, stat;
   uerr_t err;

   /* URL is like a box of chocolates -- you never know
      what you're gonna get. It's best to be safe. */
   host = nmalloc(strlen(url) + 1);
   odir = nmalloc(strlen(url) + 1);
   err = parseurl(url, host, &port, odir);
   if (err != URLHTTP && err != URLFTP)
   {
      free(odir);
      return NULL;
   }
   /* Add dir_prefix or hostname (whichever is required) to the
      beginning of dir */
   if (add_hostdir)
   {
      /* Checking for a host's true name (or at least a consistent
	 name for saving to directory) should be done here, reusing
	 the hhash, if possible. */
      if (!simple_check)
      {
	 nhost = realhost(host);
	 free(host);
	 host = nhost;
      }
      dir = nmalloc(strlen(host) + strlen(odir) + 1);
      sprintf(dir, "%s%s", host, odir);
   }
   else
   {
      dir = nmalloc(strlen(dir_prefix) + strlen(odir) + 1);
      sprintf(dir, "%s%s", dir_prefix, odir);
   }
   free(host);
   free(odir);
   for (pos = strlen(dir); pos && dir[pos] != '/'; pos--)
      if (dir[pos] == '?')
	 dir[pos] = 0; /* This takes care of CGI buggers */
   
   /* If dir is just a hostname, return index.html preceded with
      appropriate directory path. This is valid only if the URL type
      is URLHTTP. */
   if (!pos)
   {
      stat = mymkdir(dir);
      if (stat == -1)
      {
	 free(dir);
	 return NULL;
      }
      res = nmalloc(strlen(dir) + 1 + 10 + 1);
      if (err == URLHTTP)
	 sprintf(res, "%s/index.html", dir);
      else
	 sprintf(res, "%s/", dir);
      return res;
   }
   else
   {
      dir[pos] = 0;
      stat = mymkdir(dir);
      if (stat == -1)
      {
	 free(dir);
	 return NULL;
      }
      dir[pos] = '/';
      if (dir[pos + 1])
	 res = nstrdup(dir);
      else
      {
	 res = nmalloc(strlen(dir) + 10 + 1);
	 strcpy(res, dir);
	 if (err == URLHTTP)
	    strcat(res, "index.html");
      }
      free(dir);
      return res;
   }
}

/* The function that takes the dirname to be created, making sure that
   missing directories are made one by one. Its behaviour should be
   similar to mkdir -p. */
int
mymkdir(char *dir)
{
   int i, stat;

   /* Make a copy of dir, to be able to write to it. Otherwise, the
      function is unsafe if called with a const char *argument. */
   dir = nstrdup(dir);
   for (i = 0; 1; ++i)
   {
      for (; dir[i] && dir[i] != '/'; i++)
	 ;
      if (dir[i])
      {
	 dir[i] = 0;
	 stat = mkdir(dir, dirmode);
	 if (stat < 0 && errno != EEXIST)
	 {
	    free(dir);
	    return -1;
	 }
	 dir[i] = '/';
      }
      else
      {
	 stat = mkdir(dir, dirmode);
	 if (stat < 0 && errno != EEXIST)
	 {
	    free(dir);
	    return -1;
	 }
	 break;
      }
   }
   free(dir);
   return 0;
}

/* Create a unique filename, corresponding to a given URL.  If
   dirstruct is turned on, it will create all the necessary
   directories, and then return the full pathname of the file.  (it is
   important, since the function will not do what you want if you
   would like just to check for a filename) */
char *
url_filename(char *url, char *fileprefix)
{
   int count, i;
   char *name, *ptr, *murl, *pref;
   
   if (recursive && dirstruct)
   {
      char *url1;
      url1 = opt_url(url, 0);
      ptr = mkstruct(url1);
      free(url1);
      return ptr ? ptr : nstrdup(url); /* Always return *something* */
   }
   else if (prefix_files)
   {
      pref = nmalloc(strlen(dir_prefix) + 1 + strlen(fileprefix) + 1);
      sprintf(pref, "%s/%s", dir_prefix, fileprefix);
      for (count = 1; !(name = unique_name(pref, count)); count++)
	 ;
      free(pref);
      return name;
   }
   for (i = strlen(url); url[i] != '/' && i; i--)
      ;
   if (!i)
      return nstrdup(url);
   /* Checking for '~' is no longer necessary, since now it's
      done the "proper" way (through location hacks). However,
      I still provide checking for //hostname */
   if (url[i] == '/' && (!url[i + 1] || url[i - 1] == '/'))
      ptr = "index.html";
   else
      ptr = url + i + 1;
   murl = NULL;
   for (i = 0; ptr[i] && ptr[i] != '?'; i++)
      ;
   /* If there is a '?' in the string, destroy it and everything
      after it (takes care of CGI nonsense) */
   if (ptr[i] == '?')
   {
      murl = nstrdup(ptr);
      murl[i] = 0;
      ptr = murl;
   }
   pref = nmalloc(strlen(dir_prefix) + 1 + strlen(ptr) + 1);
   sprintf(pref, "%s/%s", dir_prefix, ptr);
   /* If we are dealing with FTP and always resting, no need for
      extensions. */
   if (urlproto(url) == URLFTP && always_rest)
   {
      if (murl)
	 free(murl);
      return pref;
   }
   for (count = 0; !(name = unique_name(pref, count)); count++)
      ;
   free(pref);
   if (murl)
      free(murl);
   return name;
}

/* How many digits in an integer? */
int
numdigit(int a)
{
   int res;

   for (res = 1; a /= 10; res++);
   return res;
}

/* Return a unique filename, given a prefix and count */
char *
unique_name(char *fileprefix, int count)
{
   char *filename;

   if (count)
   {
      filename = nmalloc(strlen(fileprefix) + numdigit(count) + 2);
      sprintf(filename, "%s.%d", fileprefix, count);
   }
   else
      filename = nstrdup(fileprefix);
   if (!exists(filename))
      return filename;
   else
   {
      free(filename);
      return NULL;
   }
}

/* Does a file exist? This is quite bad implementation, since it will
   return that a file does not exist if, e.g., the directory is
   unreadable. I don't mind it too much currently, though.  The proper
   way should, of course, be to check errno for ENOENT, but what about
   missing directories in the path? */
int
exists(char *filename)
{
   struct stat buf;
   return stat(filename, &buf) ? 0 : 1;
}

/* Construct an absolute URL, given a (possibly) relative one.  This
   is more tricky than it might seem, but it works. */
char *
construct(char *url, char *sub, int subsize, int no_proto)
{
   int i, fl;
   char *constr, *t;

   t = NULL;
   if (no_proto)
   {
      if (*sub != '/')
      {
	 for (i = strlen(url); i && url[i] != '/'; i--);
	 if (!i || (url[i] == url[i - 1]))
	 {
	    int l;
	    t = nmalloc((l = strlen(url)) + 2);
	    strcpy(t, url);
	    t[l] = '/';
	    t[l + 1] = 0;
	    url = t;
	    i = l;
	 }
	 constr = nmalloc(i + 1 + subsize + 1);
	 strncpy(constr, url, i + 1);
	 constr[i + 1] = 0;
	 strncat(constr, sub, subsize);
      }
      else
      {
	 i = 0;
	 do
	 {
	    for (; url[i] && url[i] != '/'; i++);
	    if (!url[i])
	       break;
	    if ((fl = (url[i] == url[i + 1] && url[i + 1] == '/')))
	       i += 2;
	 } while (fl);
	 if (!url[i])
	 {
	    int l;
	    t = nmalloc((l = strlen(url)) + 2);
	    strcpy(t, url);
	    t[l] = '/';
	    t[l + 1] = 0;
	    url = t;
	 }
	 constr = nmalloc(i + 1 + subsize + 1);
	 strncpy(constr, url, i);
	 constr[i] = 0;
	 strncat(constr + i, sub, subsize);
	 constr[i + subsize] = 0;
      }
   }
   else
   {
      constr = nmalloc(subsize + 1);
      strncpy(constr, sub, subsize);
      constr[subsize] = 0;
   }
   if (t)
      free(t);
   return constr;
}

/* Remove the occurence of /anything/../ from string s.  NOTE: things
   are not that simple, because such scheme fails when confronted with
   dir/a/b/c/../../d (which should yield dir/a/d, because it tends to
   do nasty things. This was fixed by avoiding the handling of /../../
   on purpose in each pass. Thus the first pass will produce
   dir/a/b/../d, and the second dir/a/d, QEF */
char *
strip_double(char *s)
{
   char *tmp, *res, *cur, *p;
   int l, nl, ft;

   for (ft = 1;; ft = 0)
   {
      res = nmalloc(strlen(s) + 1);
      res[0] = 0;
      for (cur = s; 1 ; cur = tmp)
      {
	 if (!(tmp = strstr(cur, "/../")))
	    break;
	 strncat(res, cur, tmp - cur);
	 if (tmp[-1] == '.' && tmp[-2] == '.' && tmp[-3] == '/')
	    continue;
	 for (p = tmp - 1, l = 0; p >= cur && *p != '/'; p--, l++);
	 nl = strlen(res) - l;
	 res[nl] = 0;
	 tmp += 3;
	 while (strstr(tmp, "/../") == tmp)
	 {
	    strcat(res, "../");
	    tmp += 3;
	 }
	 ++tmp;
      }
      strcat(res, cur);
      if (!ft)
	 free(s);
      if (!strstr(res, "/../"))
	 break;
      s = res;
   }
   return res;
}

/* When /./ encountered in s, remove ./. Things are much simpler here
   than in previous example because a/././././d is always a/d (number
   of consecutive dots is not important) */
char *
strip_dot(char *s)
{
   char *tmp, *res, *cur;
   int ft;

   for (ft = 1;; ft = 0)
   {
      res = nmalloc(strlen(s) + 1);
      res[0] = 0;
      for (cur = s; 1 ; cur = tmp + 3)
      {
	 tmp = strstr(cur, "/./");
	 if (!tmp)
	    break;
	 strncat(res, cur, tmp - cur + 1);
      }
      strcat(res, cur);
      if (!ft)
	 free(s);
      else
	 ft = 0;
      if (!strstr(res, "/./"))
	 break;
      s = res;
   }
   return res;
}

/* Lousy coding, but it works... not too slow, either (speed is not
   critical here, I believe. If flag is set, call realname,
   otherwise just dot-clean the URL. */
char *
opt_url(char *url, int fl)
{
   char *s1, *s2, *host, *dir, *tmp, *res;
   int i, port;
   uerr_t err, proto;
   
   s1 = strip_dot(url);
   s2 = strip_double(s1);
   free(s1);
   if (!fl)
      return s2;
   host = nmalloc(strlen(s2) + 1);
   dir = nmalloc(strlen(s2) + 1);
   err = parseurl(s2, host, &port, dir);
   if (err != URLHTTP && err != URLFTP)
   {
      free(host);
      free(dir);
      return s2;
   }
   free(s2);
   tmp = realhost(host);
   free(host);
   host = tmp;
   proto = urlproto(url);
   for (i = 0; sup_protos[i].name; i++)
      if (sup_protos[i].ind == proto)
	 break;
   res = nmalloc(strlen(sup_protos[i].name) + strlen(host)
		 + 1 + 6 + strlen(dir) + 1);
   sprintf(res, "%s%s:%d%s", sup_protos[i].name, host, port, dir);
   return res;
}

/* Load file to memory, return the malloc-ed buffer, and the file size */
uerr_t
load_file(FILE *fp, char **buf, long *nread)
{
   *nread = 0;
   *buf = NULL;
   while (!feof(fp))
   {
      *buf = nrealloc(*buf, *nread + FILE_BUFFER_SIZE);
      *nread += fread(*buf + *nread, sizeof(char), FILE_BUFFER_SIZE, fp);
      if (!(*buf))
	 return NOTENOUGHMEM;
   }
   return RETROK;
}

/* Create hash table, general */
void
create_hash(char ***hash, char **url)
{
   if (!*hash)
   {
      if (url)
      {
	 int i;
	 for (i = 0; url[i]; i++);
	 *hash = nmalloc((i + 1) * sizeof(char *));
	 for (i = 0; url[i]; i++)
	    (*hash)[i] = nstrdup(url[i]);
	 (*hash)[i] = NULL;
      }
      else
      {
	 *hash = nmalloc(sizeof(char *));
	 **hash = NULL;
      }
   }
}

/* Is there an entry in hash table (general)? */
int
in_hash(char **hash, char *s)
{
   for (; *hash; hash++)
      if (!strcmp(s, *hash))
	 return 1;
   return 0;
}

/* Add an entry to the hash table (general) */
void
add_hash(char ***hash, char *s)
{
   int i;
   
   for (i = 0; (*hash)[i]; i++);
   *hash = nrealloc(*hash, (i + 2) * sizeof(char *));
   (*hash)[i] = nstrdup(s);
   (*hash)[i + 1] = NULL;
#ifdef DEBUG
   if (debug)
   {
      fprintf(lfile, "Stored %s as %s in hash table.\n", s, (*hash)[i]);
      fflush(lfile);
   }
#endif
}

/* Determine whether a URL is acceptable to be followed,
   according to a list of suffixes to accept/reject */
int
acceptable(char *constr)
{
   if (accepts)
   {
      if (rejects)
	 return in_acclist(accepts, constr) && !in_acclist(rejects, constr);
      else
	 return in_acclist(accepts, constr);
   }
   else if (rejects)
      return !in_acclist(rejects, constr);
   return 1;
}

/* Is a suffix in acceptance list? */
int
in_acclist(char **accepts, char *constr)
{
   int res;
   char *suf;

   suf = suffix(constr);
   if (!suf)
      return 0;
   res = matches(accepts, suf);
   free(suf);
   return res;
}

/* Return the malloc-ed suffix of a filename */
char *
suffix(char *s)
{
   int i;
   
   for (i = strlen(s); i && s[i] != '/' && s[i] != '.'; i--);
   if (s[i++] == '.')
      return nstrdup(s + i);
   else
      return NULL;
}

/* Does a string match any element of the list? */
int
matches(char **list, char *what)
{
   int i;

   for (i = 0; list[i]; i++)
      if (!strcmp(list[i], what))
	 return 1;
   return 0;
}

/* Return the list of accepting/rejecting, given a
   comma-separated list. Strtok is used for tokenization */
char **
fillacclist(char *optstr)
{
   char *tok, **res, *tmp;
   int i, j, l;
   
   if (!optstr || !strlen(optstr))
      return NULL;
   /* Strip spaces from optstr. */
   l = strlen(optstr);
   tmp = nmalloc(l + 1);
   for (i = j = 0; i < l; i++)
      if (!isspace(optstr[i]))
	 tmp[j++] = optstr[i];
   tmp[j] = 0;
   optstr = tmp;
   if (!strchr(optstr, ','))
   {
      res = nmalloc(2 * sizeof(char *));
      res[0] = nstrdup(optstr);
      res[1] = NULL;
      free(optstr);
      return res;
   }
   tok = strtok(optstr, ",");
   res = nmalloc(2 * sizeof(char *));
   res[0] = nstrdup(tok);
   res[1] = NULL;
   for (i = 1; (tok = strtok(NULL, ",")); i++)
   {
      res = nrealloc(res, (i + 2) * sizeof(char *));
      res[i] = nstrdup(tok);
      res[i + 1] = NULL;
   }
   free(optstr);
   return res;
}

/* Free acceptance and rejectance lists */
void
freelists(void)
{
   int i;

   if (accepts)
   {
      for (i = 0; accepts[i]; i++)
	 free(accepts[i]);
      free(accepts);
   }
   if (rejects)
   {
      for (i = 0; rejects[i]; i++)
	 free(rejects[i]);
      free(rejects);
   }
}

/* Returns proxy host address, according to protocol. */
char *
getproxy(uerr_t proto)
{
   if (proto == URLHTTP)
      return http_proxy ? http_proxy : getenv("http_proxy");
   else if (proto == URLFTP)
      return ftp_proxy ? ftp_proxy : getenv("ftp_proxy");
   else
      return NULL;
}

